@@ -49,9 +49,13 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <ctype.h>
+#ifdef HAVE_SYSLOG_H
 #include <syslog.h>
+#endif
 #include <time.h>
+#ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
+#endif
 #include <locale.h>
 
 /* This makes hash table safer when debugging */
@@ -338,8 +342,14 @@ gconf_server_load_sources(void)
 
   if (addresses == NULL)
     {      
+#ifndef G_OS_WIN32
+      const char *home = g_get_home_dir ();
+#else
+      const char *home = _gconf_win32_get_home_dir ();
+#endif
+
       /* Try using the default address xml:readwrite:$(HOME)/.gconf */
-      addresses = g_slist_append(addresses, g_strconcat("xml:readwrite:", g_get_home_dir(), "/.gconf", NULL));
+      addresses = g_slist_append(addresses, g_strconcat("xml:readwrite:", home, "/.gconf", NULL));
 
       gconf_log(GCL_DEBUG, _("No configuration files found, trying to use the default config source `%s'"), (char *)addresses->data);
     }
@@ -349,7 +359,7 @@ gconf_server_load_sources(void)
       /* We want to stay alive but do nothing, because otherwise every
          request would result in another failed gconfd being spawned.  
       */
-      gconf_log(GCL_ERR, _("No configuration sources in the source path, configuration won't be saved; edit %s"), GCONF_CONFDIR"/path");
+      gconf_log(GCL_ERR, _("No configuration sources in the source path, configuration won't be saved; edit %s%s"), GCONF_CONFDIR, "/path");
       /* don't request error since there aren't any addresses */
       sources = gconf_sources_new_from_addresses(NULL, NULL);
 
@@ -369,7 +379,7 @@ gconf_server_load_sources(void)
           error = NULL;
         }
       
-      g_slist_free(addresses);
+      gconf_address_list_free(addresses);
 
       g_assert(sources != NULL);
 
@@ -413,19 +423,30 @@ signal_handler (int signo)
   switch (signo) {
     /* Fast cleanup only */
   case SIGSEGV:
+#ifdef SIGBUS
   case SIGBUS:
+#endif
   case SIGILL:
     enter_shutdown ();
+#ifndef G_OS_WIN32
     gconf_log (GCL_ERR,
                _("Received signal %d, dumping core. Please report a GConf bug."),
                signo);
     if (g_getenv ("DISPLAY"))
       gconf_handle_segv (signo);
+#else
+    gconf_log (GCL_ERR,
+               _("Received signal %d. Please report a GConf bug."),
+               signo);
+    gconf_handle_segv (signo);
+#endif
     abort ();
     break;
 
   case SIGFPE:
+#ifdef SIGPIPE
   case SIGPIPE:
+#endif
     /* Go ahead and try the full cleanup on these,
      * though it could well not work out very well.
      */
@@ -457,21 +478,28 @@ signal_handler (int signo)
       gconf_main_quit ();
     break;
 
+#ifdef SIGHUP
   case SIGHUP:
     --in_fatal;
 
     /* reload sources during next periodic_cleanup() */
     need_db_reload = TRUE;
     break;
+#endif
 
+#ifdef SIGUSR1
   case SIGUSR1:
     --in_fatal;
     
     /* it'd be nice to log a message here but it's not very safe, so */
     gconf_log_debug_messages = !gconf_log_debug_messages;
     break;
+#endif
     
   default:
+#ifndef HAVE_SIGACTION
+    signal (signo, signal_handler);
+#endif
     break;
   }
 }
@@ -525,6 +553,7 @@ log_handler (const gchar   *log_domain,
 static gboolean
 test_safe_tmp_dir (const char *dirname)
 {
+#ifndef G_OS_WIN32
   struct stat statbuf;
   int fd;
 
@@ -544,14 +573,14 @@ test_safe_tmp_dir (const char *dirname)
       return FALSE;
     }
   close (fd);
-  
+
   if (statbuf.st_uid != getuid ())
     {
       gconf_log (GCL_WARNING, _("Owner of %s is not the current user"),
                  dirname);
       return FALSE;
     }
-  
+
   if ((statbuf.st_mode & (S_IRWXG|S_IRWXO)) ||
       !S_ISDIR (statbuf.st_mode))
     {
@@ -559,6 +588,12 @@ test_safe_tmp_dir (const char *dirname)
                  (unsigned long) statbuf.st_mode & 07777, dirname);
       return FALSE;
     }
+#else
+  /* FIXME: We can't get any useful information about the actual
+   * protection for the directory using stat(). We must use the Win32
+   * API to check the owner and permissions (ACL). Later.
+   */
+#endif
   
   return TRUE;
 }
@@ -566,8 +601,10 @@ test_safe_tmp_dir (const char *dirname)
 int 
 main(int argc, char** argv)
 {
+#ifdef HAVE_SIGACTION
   struct sigaction act;
   sigset_t empty_mask;
+#endif
   CORBA_Environment ev;
   CORBA_ORB orb;
   gchar* logname;
@@ -596,13 +633,13 @@ main(int argc, char** argv)
   if (chdir ("/") < 0)
     {
        g_printerr ("Could not change to root directory: %s\n",
-		g_strerror (errno));
+		   g_strerror (errno));
        exit (1);
     }
 
   if (!g_getenv ("GCONF_DEBUG_OUTPUT"))
     {
-      dev_null_fd = open ("/dev/null", O_RDWR);
+      dev_null_fd = open (DEV_NULL, O_RDWR);
       if (dev_null_fd >= 0)
         {
 	  dup2 (dev_null_fd, 0);
@@ -623,7 +660,9 @@ main(int argc, char** argv)
   username = g_get_user_name();
   logname = g_strdup_printf("gconfd (%s-%u)", username, (guint)getpid());
 
+#ifdef HAVE_SYSLOG_H
   openlog (logname, LOG_NDELAY, LOG_USER);
+#endif
 
   g_log_set_handler (NULL, G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
                      log_handler, NULL);
@@ -645,21 +684,38 @@ main(int argc, char** argv)
 #endif
   
   /* Session setup */
+#ifdef HAVE_SIGACTION
   sigemptyset (&empty_mask);
   act.sa_handler = signal_handler;
   act.sa_mask    = empty_mask;
   act.sa_flags   = 0;
-  sigaction (SIGTERM,  &act, 0);
-  sigaction (SIGILL,  &act, 0);
-  sigaction (SIGBUS,  &act, 0);
-  sigaction (SIGFPE,  &act, 0);
-  sigaction (SIGHUP,  &act, 0);
-  sigaction (SIGSEGV, &act, 0);
-  sigaction (SIGABRT, &act, 0);
-  sigaction (SIGUSR1,  &act, 0);
-  
+  sigaction (SIGTERM,  &act, NULL);
+  sigaction (SIGILL,  &act, NULL);
+  sigaction (SIGBUS,  &act, NULL);
+  sigaction (SIGFPE,  &act, NULL);
+  sigaction (SIGHUP,  &act, NULL);
+  sigaction (SIGSEGV, &act, NULL);
+  sigaction (SIGABRT, &act, NULL);
+  sigaction (SIGUSR1,  &act, NULL);
+
   act.sa_handler = SIG_IGN;
-  sigaction (SIGINT, &act, 0);
+  sigaction (SIGINT, &act, NULL);
+#else
+  signal (SIGTERM, signal_handler);
+  signal (SIGILL,  signal_handler);
+#ifdef SIGBUS
+  signal (SIGBUS,  signal_handler);
+#endif
+  signal (SIGFPE,  signal_handler);
+#ifdef SIGHUP
+  signal (SIGHUP,  signal_handler);
+#endif
+  signal (SIGSEGV, signal_handler);
+  signal (SIGABRT, signal_handler);
+#ifdef SIGUSR1
+  signal (SIGUSR1,  signal_handler);
+#endif
+#endif
 
   CORBA_exception_init(&ev);
 
@@ -689,10 +745,10 @@ main(int argc, char** argv)
   gconfd_dir = gconf_get_daemon_dir ();
   lock_dir = gconf_get_lock_dir ();
   
-  if (mkdir (gconfd_dir, 0700) < 0 && errno != EEXIST)
+  if (g_mkdir (gconfd_dir, 0700) < 0 && errno != EEXIST)
     gconf_log (GCL_WARNING, _("Failed to create %s: %s"),
                gconfd_dir, g_strerror (errno));
-
+  
   if (!test_safe_tmp_dir (gconfd_dir))
     {
       err = g_error_new (GCONF_ERROR,
@@ -708,6 +764,9 @@ main(int argc, char** argv)
       daemon_lock = gconf_get_lock (lock_dir, &err);
     }
 
+  g_free (gconfd_dir);
+  g_free (lock_dir);
+
   if (daemon_lock != NULL)
     {
       /* This loads backends and so on. It needs to be done before
@@ -726,9 +785,9 @@ main(int argc, char** argv)
       char buf[1] = { 'g' };
       if (write (write_byte_fd, buf, 1) != 1)
         {
-          gconf_log (GCL_ERR, _("Failed to write byte to pipe fd %d so client program may hang: %s"), write_byte_fd, g_strerror (errno));
+          gconf_log (GCL_ERR, _("Failed to write byte to pipe file descriptor %d so client program may hang: %s"), write_byte_fd, g_strerror (errno));
         }
-
+      
       close (write_byte_fd);
     }
   
@@ -806,9 +865,11 @@ periodic_cleanup_timeout(gpointer data)
       gconf_log (GCL_INFO, _("SIGHUP received, reloading all databases"));
 
       need_db_reload = FALSE;
+      logfile_save ();
       shutdown_databases ();
       init_databases ();
-      gconf_server_load_sources ();		  
+      gconf_server_load_sources ();
+      logfile_read ();
     }
   
   gconf_log (GCL_DEBUG, "Performing periodic cleanup, expiring cache cruft");
@@ -1021,6 +1082,12 @@ drop_old_databases(void)
     {
       GConfDatabase* db = tmp_list->data;
 
+      if (db == default_db)
+	{
+	  tmp_list = g_list_next (tmp_list);
+	  continue;
+	}
+
       /* Drop any listeners whose clients are gone. */
       gconf_database_drop_dead_listeners (db);
       
@@ -1083,8 +1150,15 @@ no_databases_in_use (void)
   /* Only the default database still open, and
    * it has no listeners
    */
-  return db_list == NULL &&
-    gconf_listeners_count (default_db->listeners) == 0;
+
+  if (db_list == NULL)
+    return TRUE;
+
+  if (db_list->next == NULL &&
+      db_list->data == default_db)
+    return gconf_listeners_count (default_db->listeners) == 0;
+
+  return FALSE;
 }
 
 void
@@ -1137,7 +1211,16 @@ gconfd_notify_other_listeners (GConfDatabase *modified_db,
 		      return;
 		    }
 
-		  cvalue = gconf_corba_value_from_gconf_value (value);
+		  if (value != NULL)
+		    {
+		      cvalue = gconf_corba_value_from_gconf_value (value);
+		      gconf_value_free (value);
+		    }
+		  else
+		    {
+		      cvalue = gconf_invalid_corba_value ();
+		    }
+
 		  gconf_database_notify_listeners (db,
 						   NULL,
 						   key,
@@ -1146,7 +1229,6 @@ gconfd_notify_other_listeners (GConfDatabase *modified_db,
 						   is_writable,
 						   FALSE);
 		  CORBA_free (cvalue);
-		  gconf_value_free (value);
 		}
 
 	      tmp2 = tmp2->next;
@@ -1312,8 +1394,14 @@ gconfd_check_in_shutdown (CORBA_Environment *ev)
 static void
 get_log_names (gchar **logdir, gchar **logfile)
 {
-  *logdir = gconf_concat_dir_and_key (g_get_home_dir (), ".gconfd");
-  *logfile = gconf_concat_dir_and_key (*logdir, "saved_state");
+#ifndef G_OS_WIN32
+      const char *home = g_get_home_dir ();
+#else
+      const char *home = _gconf_win32_get_home_dir ();
+#endif
+
+  *logdir = g_build_filename (home, ".gconfd", NULL);
+  *logfile = g_build_filename (*logdir, "saved_state", NULL);
 }
 
 static void close_append_handle (void);
@@ -1341,18 +1429,18 @@ open_append_handle (GError **err)
 
       get_log_names (&logdir, &logfile);
       
-      mkdir (logdir, 0700); /* ignore failure, we'll catch failures
-                             * that matter on open()
-                             */
+      g_mkdir (logdir, 0700); /* ignore failure, we'll catch failures
+			       * that matter on open()
+			       */
       
-      append_handle = fopen (logfile, "a");
+      append_handle = g_fopen (logfile, "a");
 
       if (append_handle == NULL)
         {
           gconf_set_error (err,
                            GCONF_ERROR_FAILED,
                            _("Failed to open gconfd logfile; won't be able to restore listeners after gconfd shutdown (%s)"),
-                           strerror (errno));
+                           g_strerror (errno));
           
           g_free (logdir);
           g_free (logfile);
@@ -1387,7 +1475,7 @@ close_append_handle (void)
       if (fclose (append_handle) < 0)
         gconf_log (GCL_WARNING,
                    _("Failed to close gconfd logfile; data may not have been properly saved (%s)"),
-                   strerror (errno));
+                   g_strerror (errno));
 
       append_handle = NULL;
 
@@ -1418,9 +1506,9 @@ logfile_save (void)
   
   get_log_names (&logdir, &logfile);
 
-  mkdir (logdir, 0700); /* ignore failure, we'll catch failures
-                         * that matter on open()
-                         */
+  g_mkdir (logdir, 0700); /* ignore failure, we'll catch failures
+			   * that matter on open()
+			   */
 
   saveme = g_string_new (NULL);
 
@@ -1443,13 +1531,13 @@ logfile_save (void)
   /* Now try saving the string to a temporary file */
   tmpfile = g_strconcat (logfile, ".tmp", NULL);
   
-  fd = open (tmpfile, O_WRONLY | O_CREAT | O_TRUNC, 0700);
+  fd = g_open (tmpfile, O_WRONLY | O_CREAT | O_TRUNC, 0700);
 
   if (fd < 0)
     {
       gconf_log (GCL_WARNING,
                  _("Could not open saved state file '%s' for writing: %s"),
-                 tmpfile, strerror (errno));
+                 tmpfile, g_strerror (errno));
       
       goto out;
     }
@@ -1463,7 +1551,7 @@ logfile_save (void)
       
       gconf_log (GCL_WARNING,
                  _("Could not write saved state file '%s' fd: %d: %s"),
-                 tmpfile, fd, strerror (errno));
+                 tmpfile, fd, g_strerror (errno));
 
       goto out;
     }
@@ -1472,7 +1560,7 @@ logfile_save (void)
     {
       gconf_log (GCL_WARNING,
                  _("Failed to close new saved state file '%s': %s"),
-                 tmpfile, strerror (errno));
+                 tmpfile, g_strerror (errno));
       goto out;
     }
 
@@ -1482,30 +1570,30 @@ logfile_save (void)
   if (gconf_file_exists (logfile))
     {
       tmpfile2 = g_strconcat (logfile, ".orig", NULL);
-      if (rename (logfile, tmpfile2) < 0)
+      if (g_rename (logfile, tmpfile2) < 0)
         {
           gconf_log (GCL_WARNING,
                      _("Could not move aside old saved state file '%s': %s"),
-                     logfile, strerror (errno));
+                     logfile, g_strerror (errno));
           goto out;
         }
     }
 
   /* Move the new saved state file into place */
-  if (rename (tmpfile, logfile) < 0)
+  if (g_rename (tmpfile, logfile) < 0)
     {
       gconf_log (GCL_WARNING,
                  _("Failed to move new save state file into place: %s"),
-                 strerror (errno));
+                 g_strerror (errno));
 
       /* Try to restore old file */
       if (tmpfile2)
         {
-          if (rename (tmpfile2, logfile) < 0)
+          if (g_rename (tmpfile2, logfile) < 0)
             {
               gconf_log (GCL_WARNING,
                          _("Failed to restore original saved state file that had been moved to '%s': %s"),
-                         tmpfile2, strerror (errno));
+                         tmpfile2, g_strerror (errno));
 
             }
         }
@@ -1515,7 +1603,7 @@ logfile_save (void)
 
   /* Get rid of original saved state file if everything succeeded */
   if (tmpfile2)
-    unlink (tmpfile2);
+    g_unlink (tmpfile2);
   
  out:
   if (saveme)
@@ -2003,63 +2091,49 @@ restore_client_foreach (gpointer key,
 }
 
 
-#ifndef HAVE_FLOCKFILE
-#  define flockfile(f) (void)1
-#  define funlockfile(f) (void)1
-#  define getc_unlocked(f) getc(f)
-#endif /* !HAVE_FLOCKFILE */
-
 static gchar*
 read_line (FILE *f)
 {
-  int c;
-  GString *str;
-  
-  str = g_string_new (NULL);
-  
-  flockfile (f);
+#define BUF_SIZE 2048
 
-  while (TRUE)
+  char  buf[BUF_SIZE] = { '\0' };
+  char *retval = NULL;
+  int   len = 0;
+
+  do
     {
-      c = getc_unlocked (f);
-
-      switch (c)
+      if (fgets (buf, BUF_SIZE, f) == NULL)
         {
-        case EOF:
           if (ferror (f))
             {
               gconf_log (GCL_ERR,
                          _("Error reading saved state file: %s"),
                          g_strerror (errno));
             }
-
-          /* FALL THRU */
-        case '\n':
-          goto done;
-          break;
-
-        default:
-          g_string_append_c (str, c);
           break;
         }
-    }
 
- done:
-  funlockfile (f);
+      len = strlen (buf);
+      if (buf[len - 1] == '\n')
+	buf[--len] = '\0';
 
-  if (str->len == 0)
-    {
-      g_string_free (str, TRUE);
-      return NULL;
-    }
-  else
-    {
-      gchar *ret;
-      
-      ret = str->str;
-      g_string_free (str, FALSE);
-      return ret;
+      if (retval == NULL)
+	{
+	  retval = g_strndup (buf, len);
+	}
+      else
+	{
+	  char *freeme = retval;
+
+	  retval = g_strconcat (retval, buf, NULL);
+	  g_free (freeme);
+	}
     }
+  while (len == BUF_SIZE - 1);
+
+  return retval;
+
+#undef BUF_SIZE
 }
 
 static void
@@ -2078,7 +2152,7 @@ logfile_read (void)
   
   get_log_names (&logdir, &logfile);
 
-  f = fopen (logfile, "r");
+  f = g_fopen (logfile, "r");
   
   if (f == NULL)
     {
@@ -2259,7 +2333,7 @@ log_client_change (const ConfigListener client,
     {
       gconf_log (GCL_WARNING,
                  _("Failed to write client add to saved state file: %s"),
-                 strerror (errno));
+                 g_strerror (errno));
       goto error;
     }
 
@@ -2267,7 +2341,7 @@ log_client_change (const ConfigListener client,
     {
       gconf_log (GCL_WARNING,
                  _("Failed to flush client add to saved state file: %s"),
-                 strerror (errno));
+                 g_strerror (errno));
       goto error;
     }
 
